package rtr;

import addr.addrIP;
import addr.addrIPv4;
import addr.addrMac;
import addr.addrPrefix;
import cfg.cfgAceslst;
import cfg.cfgAll;
import cfg.cfgPlymp;
import cfg.cfgPrfxlst;
import cfg.cfgProxy;
import cfg.cfgRoump;
import cfg.cfgRouplc;
import cfg.cfgRtr;
import cfg.cfgVrf;
import ifc.ifcDot1ah;
import ip.ipCor4;
import ip.ipCor6;
import ip.ipFwd;
import ip.ipFwdIface;
import ip.ipRtr;
import java.util.List;
import java.util.ArrayList;
import pipe.pipeLine;
import pipe.pipeSide;
import prt.prtGenConn;
import prt.prtServS;
import prt.prtTcp;
import tab.tabListing;
import tab.tabAceslstN;
import tab.tabGen;
import tab.tabLabel;
import tab.tabLabelNtry;
import tab.tabPlcmapN;
import tab.tabPrfxlstN;
import tab.tabRoute;
import tab.tabRouteEntry;
import tab.tabRtrmapN;
import tab.tabRtrplcN;
import user.userFormat;
import user.userHelping;
import util.bits;
import util.cmds;
import util.debugger;
import util.logger;
import util.notifier;

/**
 * border gateway protocol (rfc4271) version 4
 *
 * @author matecsaba
 */
public class rtrBgp extends ipRtr implements prtServS, Runnable {

    /**
     * port to use
     */
    public final static int port = 179;

    /**
     * local as number
     */
    public int localAs;

    /**
     * address families
     */
    public int addrFams;

    /**
     * router id
     */
    public addrIPv4 routerID;

    /**
     * scan time interval
     */
    public int scanTime;

    /**
     * restart time
     */
    public int restartTime;

    /**
     * external distance
     */
    public int distantExt;

    /**
     * internal distance
     */
    public int distantInt;

    /**
     * local distance
     */
    public int distantLoc;

    /**
     * update groups
     */
    public List<rtrBgpGroup> groups;

    /**
     * listen template
     */
    public rtrBgpTemp lstnTmp;

    /**
     * listen access list
     */
    public tabListing<tabAceslstN<addrIP>, addrIP> lstnAcl;

    /**
     * list of dynamic neighbors
     */
    protected tabGen<rtrBgpNeigh> lstnNei;

    /**
     * route type
     */
    protected tabRouteEntry.routeType rouTyp;

    /**
     * unicast afi
     */
    protected int afiUni;

    /**
     * labeled unicast afi
     */
    protected int afiLab;

    /**
     * multicast afi
     */
    protected int afiMlt;

    /**
     * flow specification afi
     */
    protected int afiFlw;

    /**
     * unicast vpn afi
     */
    protected int afiVpnU;

    /**
     * multicast vpn afi
     */
    protected int afiVpnM;

    /**
     * flowspec vpn afi
     */
    protected int afiVpnF;

    /**
     * vpls afi
     */
    protected int afiVpls;

    /**
     * evpn afi
     */
    protected int afiEvpn;

    /**
     * mdt afi
     */
    protected int afiMdt;

    /**
     * router number
     */
    protected int rtrNum;

    /**
     * next hop tracking route map
     */
    protected tabListing<tabRtrmapN, addrIP> nhtRoumap;

    /**
     * next hop tracking route policy
     */
    protected tabListing<tabRtrplcN, addrIP> nhtRouplc;

    /**
     * next hop tracking policy map
     */
    protected tabListing<tabPrfxlstN, addrIP> nhtPfxlst;

    /**
     * flow specification
     */
    protected tabListing<tabPlcmapN, addrIP> flowSpec;

    /**
     * list of rpkis
     */
    protected tabGen<rtrBgpRpki> rpkis;

    /**
     * the rpki routes
     */
    public tabRoute<addrIP> computedRpki = new tabRoute<addrIP>("rx");

    /**
     * the multicast routes
     */
    public tabRoute<addrIP> computedMlt = new tabRoute<addrIP>("rx");

    /**
     * the flowspec routes
     */
    public tabRoute<addrIP> computedFlw = new tabRoute<addrIP>("rx");

    /**
     * the vpnuni routes
     */
    public tabRoute<addrIP> computedVpnU = new tabRoute<addrIP>("rx");

    /**
     * the vpnmlt routes
     */
    public tabRoute<addrIP> computedVpnM = new tabRoute<addrIP>("rx");

    /**
     * the vpnflw routes
     */
    public tabRoute<addrIP> computedVpnF = new tabRoute<addrIP>("rx");

    /**
     * the vpls routes
     */
    public tabRoute<addrIP> computedVpls = new tabRoute<addrIP>("rx");

    /**
     * the evpn routes
     */
    public tabRoute<addrIP> computedEvpn = new tabRoute<addrIP>("rx");

    /**
     * the mdt routes
     */
    public tabRoute<addrIP> computedMdt = new tabRoute<addrIP>("rx");

    /**
     * the unicast routes
     */
    public tabRoute<addrIP> changedUni = new tabRoute<addrIP>("rx");

    /**
     * the multicast routes
     */
    public tabRoute<addrIP> changedMlt = new tabRoute<addrIP>("rx");

    /**
     * the flowspec routes
     */
    public tabRoute<addrIP> changedFlw = new tabRoute<addrIP>("rx");

    /**
     * the vpnuni routes
     */
    public tabRoute<addrIP> changedVpnU = new tabRoute<addrIP>("rx");

    /**
     * the vpnmlt routes
     */
    public tabRoute<addrIP> changedVpnM = new tabRoute<addrIP>("rx");

    /**
     * the vpnflw routes
     */
    public tabRoute<addrIP> changedVpnF = new tabRoute<addrIP>("rx");

    /**
     * the vpls routes
     */
    public tabRoute<addrIP> changedVpls = new tabRoute<addrIP>("rx");

    /**
     * the evpn routes
     */
    public tabRoute<addrIP> changedEvpn = new tabRoute<addrIP>("rx");

    /**
     * the mdt routes
     */
    public tabRoute<addrIP> changedMdt = new tabRoute<addrIP>("rx");

    /**
     * incremental limit
     */
    public int incrLimit;

    /**
     * conquer bestpath
     */
    public boolean conquer;

    /**
     * list of monitors
     */
    protected tabGen<rtrBgpMon> mons;

    /**
     * list of dumps
     */
    protected tabGen<rtrBgpMrt> dmps;

    /**
     * list of neighbors
     */
    protected tabGen<rtrBgpNeigh> neighs;

    /**
     * list of templates
     */
    protected tabGen<rtrBgpTemp> temps;

    /**
     * list of vrfs
     */
    protected tabGen<rtrBgpVrf> vrfs;

    /**
     * list of vpls
     */
    protected tabGen<rtrBgpVpls> vpls;

    /**
     * list of evpns
     */
    protected tabGen<rtrBgpEvpn> evpn;

    /**
     * evpn receiver
     */
    protected rtrBgpEvpnPbb evpnRcv;

    /**
     * evpn unicast label
     */
    protected tabLabelNtry evpnUni;

    /**
     * evpn multicast label
     */
    protected tabLabelNtry evpnMul;

    /**
     * full compute last
     */
    public long fullLast;

    /**
     * incremental compute last
     */
    public long incrLast;

    /**
     * full compute count
     */
    public int fullCount;

    /**
     * incremental compute count
     */
    public int incrCount;

    /**
     * full compute time
     */
    public int fullTime;

    /**
     * incremental compute time
     */
    public int incrTime;

    /**
     * changed prefixes current
     */
    public int changedCur;

    /**
     * changed prefixes total
     */
    public long changedTot;

    /**
     * the tcp protocol
     */
    public final prtTcp tcpCore;

    /**
     * the forwarder protocol
     */
    public final ipFwd fwdCore;

    /**
     * the forwarder vrf
     */
    protected cfgVrf vrfCore;

    /**
     * notifier for table computation
     */
    protected notifier compute = new notifier();

    /**
     * comutation round
     */
    protected int compRound;

    /**
     * need full round
     */
    protected boolean needFull;

    private boolean oldAggr;

    private boolean need2run;

    /**
     * create bgp process
     *
     * @param forwarder forwarder to update
     * @param vrfcfg vrf config to use
     * @param protoT tcp protocol to use
     * @param id process id
     */
    public rtrBgp(ipFwd forwarder, cfgVrf vrfcfg, prtTcp protoT, int id) {
        if (debugger.rtrBgpEvnt) {
            logger.debug("startup");
        }
        vrfCore = vrfcfg;
        fwdCore = forwarder;
        tcpCore = protoT;
        vrfs = new tabGen<rtrBgpVrf>();
        vpls = new tabGen<rtrBgpVpls>();
        evpn = new tabGen<rtrBgpEvpn>();
        evpnUni = tabLabel.allocate(10);
        evpnMul = tabLabel.allocate(10);
        evpnRcv = new rtrBgpEvpnPbb(this);
        evpnUni.setFwdPwe(10, fwdCore, evpnRcv, 0, null);
        evpnMul.setFwdPwe(10, fwdCore, evpnRcv, 0, null);
        routerID = new addrIPv4();
        addrFams = rtrBgpParam.mskUni;
        rtrNum = id;
        switch (fwdCore.ipVersion) {
            case ipCor4.protocolVersion:
                rouTyp = tabRouteEntry.routeType.bgp4;
                afiUni = rtrBgpUtil.safiIp4uni;
                afiLab = rtrBgpUtil.safiIp4lab;
                afiMlt = rtrBgpUtil.safiIp4multi;
                afiFlw = rtrBgpUtil.safiIp4flow;
                afiVpnU = rtrBgpUtil.safiIp4vpnU;
                afiVpnM = rtrBgpUtil.safiIp4vpnM;
                afiVpnF = rtrBgpUtil.safiIp4vpnF;
                afiVpls = rtrBgpUtil.safiVpls4;
                afiEvpn = rtrBgpUtil.safiEvpn4;
                afiMdt = rtrBgpUtil.safiIp4mdt;
                break;
            case ipCor6.protocolVersion:
                rouTyp = tabRouteEntry.routeType.bgp6;
                afiUni = rtrBgpUtil.safiIp6uni;
                afiLab = rtrBgpUtil.safiIp6lab;
                afiMlt = rtrBgpUtil.safiIp6multi;
                afiFlw = rtrBgpUtil.safiIp6flow;
                afiVpnU = rtrBgpUtil.safiIp6vpnU;
                afiVpnM = rtrBgpUtil.safiIp6vpnM;
                afiVpnF = rtrBgpUtil.safiIp6vpnF;
                afiVpls = rtrBgpUtil.safiVpls6;
                afiEvpn = rtrBgpUtil.safiEvpn6;
                afiMdt = rtrBgpUtil.safiIp6mdt;
                break;
        }
        incrLimit = 1000;
        conquer = false;
        scanTime = 1000;
        restartTime = 60 * 1000;
        distantExt = 20;
        distantInt = 200;
        distantLoc = 200;
        lstnNei = new tabGen<rtrBgpNeigh>();
        neighs = new tabGen<rtrBgpNeigh>();
        mons = new tabGen<rtrBgpMon>();
        dmps = new tabGen<rtrBgpMrt>();
        rpkis = new tabGen<rtrBgpRpki>();
        temps = new tabGen<rtrBgpTemp>();
        routerComputed = new tabRoute<addrIP>("rx");
        needFull = true;
        compRound = 1;
        routerCreateComputed();
        need2run = true;
        new Thread(this).start();
        fwdCore.routerAdd(this, rouTyp, id);
    }

    /**
     * convert safi to mask
     *
     * @param safi safi
     * @return mask
     */
    public int safi2mask(int safi) {
        if (safi == afiUni) {
            return rtrBgpParam.mskUni;
        }
        if (safi == afiLab) {
            return rtrBgpParam.mskLab;
        }
        if (safi == afiMlt) {
            return rtrBgpParam.mskMlt;
        }
        if (safi == afiFlw) {
            return rtrBgpParam.mskFlw;
        }
        if (safi == afiVpnU) {
            return rtrBgpParam.mskVpnU;
        }
        if (safi == afiVpnM) {
            return rtrBgpParam.mskVpnM;
        }
        if (safi == afiVpnF) {
            return rtrBgpParam.mskVpnF;
        }
        if (safi == afiVpls) {
            return rtrBgpParam.mskVpls;
        }
        if (safi == afiEvpn) {
            return rtrBgpParam.mskEvpn;
        }
        if (safi == afiMdt) {
            return rtrBgpParam.mskMdt;
        }
        return -1;
    }

    /**
     * convert mask to safi
     *
     * @param mask mask
     * @return safi
     */
    public int mask2safi(int mask) {
        switch (mask) {
            case rtrBgpParam.mskUni:
                return afiUni;
            case rtrBgpParam.mskLab:
                return afiLab;
            case rtrBgpParam.mskMlt:
                return afiMlt;
            case rtrBgpParam.mskFlw:
                return afiFlw;
            case rtrBgpParam.mskVpnU:
                return afiVpnU;
            case rtrBgpParam.mskVpnM:
                return afiVpnM;
            case rtrBgpParam.mskVpnF:
                return afiVpnF;
            case rtrBgpParam.mskVpls:
                return afiVpls;
            case rtrBgpParam.mskEvpn:
                return afiEvpn;
            case rtrBgpParam.mskMdt:
                return afiMdt;
            default:
                return -1;
        }
    }

    /**
     * mask to list
     *
     * @param mask mask
     * @return list
     */
    public List<Integer> mask2list(int mask) {
        List<Integer> safis = new ArrayList<Integer>();
        if ((mask & rtrBgpParam.mskUni) != 0) {
            safis.add(afiUni);
        }
        if ((mask & rtrBgpParam.mskLab) != 0) {
            safis.add(afiLab);
        }
        if ((mask & rtrBgpParam.mskMlt) != 0) {
            safis.add(afiMlt);
        }
        if ((mask & rtrBgpParam.mskFlw) != 0) {
            safis.add(afiFlw);
        }
        if ((mask & rtrBgpParam.mskVpnU) != 0) {
            safis.add(afiVpnU);
        }
        if ((mask & rtrBgpParam.mskVpnM) != 0) {
            safis.add(afiVpnM);
        }
        if ((mask & rtrBgpParam.mskVpnF) != 0) {
            safis.add(afiVpnF);
        }
        if ((mask & rtrBgpParam.mskVpls) != 0) {
            safis.add(afiVpls);
        }
        if ((mask & rtrBgpParam.mskEvpn) != 0) {
            safis.add(afiEvpn);
        }
        if ((mask & rtrBgpParam.mskMdt) != 0) {
            safis.add(afiMdt);
        }
        return safis;
    }

    public void run() {
        for (;;) {
            if (compute.missedWakes() > 0) {
                bits.sleep(scanTime);
            } else {
                compute.sleep(0);
            }
            if (!need2run) {
                break;
            }
            routerCreateComputed();
        }
    }

    public void closedInterface(ipFwdIface ifc) {
    }

    public boolean streamAccept(pipeSide pipe, prtGenConn id) {
        if (lstnAcl == null) {
            return true;
        }
        if (lstnTmp == null) {
            return true;
        }
        if (!lstnAcl.matches(id)) {
            return true;
        }
        rtrBgpNeigh ntry = new rtrBgpNeigh(this);
        ntry.peerAddr = id.peerAddr.copyBytes();
        ntry.localIfc = id.iface;
        ntry.localAddr = id.iface.addr.copyBytes();
        if (neighs.find(ntry) != null) {
            return true;
        }
        ntry.copyFrom(lstnTmp);
        ntry.updatePeer();
        rtrBgpNeigh res = lstnNei.put(ntry);
        if (res != null) {
            res.socketMode = 5;
            res.stopNow();
        }
        ntry.conn = new rtrBgpSpeak(this, ntry, pipe);
        ntry.socketMode = 4;
        ntry.startNow();
        return false;
    }

    public boolean streamForceBlock() {
        return true;
    }

    public void routerRedistChanged() {
        needFull = true;
        compute.wakeup();
    }

    public void routerOthersChanged() {
        if ((nhtRoumap == null) && (nhtRouplc == null) && (nhtPfxlst == null)) {
            return;
        }
        compute.wakeup();
    }

    private void computeFull() {
        long tim = bits.getTime();
        if (debugger.rtrBgpIncr) {
            logger.debug("bestpath for everything");
        }
        changedUni = new tabRoute<addrIP>("bgp");
        changedMlt = new tabRoute<addrIP>("bgp");
        changedFlw = new tabRoute<addrIP>("bgp");
        changedVpnU = new tabRoute<addrIP>("bgp");
        changedVpnM = new tabRoute<addrIP>("bgp");
        changedVpnF = new tabRoute<addrIP>("bgp");
        changedVpls = new tabRoute<addrIP>("bgp");
        changedEvpn = new tabRoute<addrIP>("bgp");
        changedMdt = new tabRoute<addrIP>("bgp");
        needFull = false;
        tabRoute<addrIP> nUni = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nMlt = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nFlw = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nVpnU = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nVpnM = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nVpnF = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nVpls = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nEvpn = new tabRoute<addrIP>("bst");
        tabRoute<addrIP> nMdt = new tabRoute<addrIP>("bst");
        if (flowSpec != null) {
            rtrBgpFlow.doAdvertise(nFlw, flowSpec, new tabRouteEntry<addrIP>(), afiUni == rtrBgpUtil.safiIp6uni, localAs);
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.doAdvertise(nVpnU, nVpnF);
        }
        nVpnM.mergeFrom(3, nVpnU, null, true);
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).doAdvertise(nVpls);
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).doAdvertise(nEvpn);
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " rpki");
        }
        computedRpki = new tabRoute<addrIP>("bgp");
        for (int i = 0; i < rpkis.size(); i++) {
            computedRpki.mergeFrom(2, rpkis.get(i).table, null, true);
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " neighbors");
        }
        groups = new ArrayList<rtrBgpGroup>();
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh nei = lstnNei.get(i);
            nei.setAccepted();
            nei.setGroup();
            nei.setValidity();
            nei.setMerge(nUni, nMlt, nFlw, nVpnU, nVpnM, nVpnF, nVpls, nEvpn, nMdt);
        }
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh nei = neighs.get(i);
            nei.setAccepted();
            nei.setGroup();
            nei.setValidity();
            nei.setMerge(nUni, nMlt, nFlw, nVpnU, nVpnM, nVpnF, nVpls, nEvpn, nMdt);
        }
        if (conquer) {
            if (debugger.rtrBgpComp) {
                logger.debug("round " + compRound + " counquer");
            }
            computeConquerTable(routerComputed, nUni);
            computeConquerTable(computedMlt, nMlt);
            computeConquerTable(computedFlw, nFlw);
            computeConquerTable(computedVpnU, nVpnU);
            computeConquerTable(computedVpnM, nVpnM);
            computeConquerTable(computedVpnF, nVpnF);
            computeConquerTable(computedVpls, nVpls);
            computeConquerTable(computedEvpn, nEvpn);
            computeConquerTable(computedMdt, nMdt);
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " groups");
        }
        for (int i = 0; i < groups.size(); i++) {
            groups.get(i).createNeeded(nUni, nMlt, nFlw, nVpnU, nVpnM, nVpnF, nVpls, nEvpn, nMdt);
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " neigroups");
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            lstnNei.get(i).setNeeded();
        }
        for (int i = 0; i < neighs.size(); i++) {
            neighs.get(i).setNeeded();
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " export");
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.doPeers(nVpnU);
        }
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).doPeers(nVpls);
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).doPeers(nEvpn);
        }
        routerComputed = nUni;
        computedMlt = nMlt;
        computedFlw = nFlw;
        computedVpnU = nVpnU;
        computedVpnM = nVpnM;
        computedVpnF = nVpnF;
        computedVpls = nVpls;
        computedEvpn = nEvpn;
        computedMdt = nMdt;
        fullLast = bits.getTime();
        fullTime = (int) (fullLast - tim);
        fullCount++;
    }

    private void computeIncrEntry(int afi, tabRouteEntry<addrIP> curr, tabRoute<addrIP> cmp) {
        if (debugger.rtrBgpIncr) {
            logger.debug("bestpath for " + tabRtrmapN.rd2string(curr.rouDst) + " " + curr.prefix + " in " + rtrBgpUtil.safi2string(afi));
        }
        tabRouteEntry<addrIP> best = null;
        if ((afi == afiUni) || (afi == afiMlt)) {
            best = routerRedisted.find(curr);
            if (best != null) {
                best = best.copyBytes();
                best.rouSrc = rtrBgpUtil.peerOriginate;
            }
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh nei = lstnNei.get(i);
            if (!nei.reachable) {
                continue;
            }
            tabRoute<addrIP> acc = nei.getAccepted(afi);
            if (acc == null) {
                needFull = true;
                continue;
            }
            tabRouteEntry<addrIP> ntry = acc.find(curr);
            if (ntry == null) {
                continue;
            }
            setValidity(ntry);
            if (best == null) {
                best = ntry;
                continue;
            }
            if (!best.isOtherBetter(ntry)) {
                continue;
            }
            best = ntry;
        }
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh nei = neighs.get(i);
            if (!nei.reachable) {
                continue;
            }
            tabRoute<addrIP> acc = nei.getAccepted(afi);
            if (acc == null) {
                needFull = true;
                continue;
            }
            tabRouteEntry<addrIP> ntry = acc.find(curr);
            if (ntry == null) {
                continue;
            }
            setValidity(ntry);
            if (best == null) {
                best = ntry;
                continue;
            }
            if (!best.isOtherBetter(ntry)) {
                continue;
            }
            best = ntry;
        }
        if (best == null) {
            cmp.del(curr);
            curr.version = compRound + 1;
            for (int i = 0; i < groups.size(); i++) {
                rtrBgpGroup grp = groups.get(i);
                tabRoute<addrIP> wil = grp.getWilling(afi);
                tabRoute<addrIP> chg = grp.getChanged(afi);
                if ((wil == null) || (chg == null)) {
                    needFull = true;
                    continue;
                }
                if (wil.del(curr)) {
                    continue;
                }
                chg.add(3, curr, false, false);
            }
            return;
        }
        best.version = compRound + 1;
        if (conquer) {
            tabRouteEntry<addrIP> res = computeConquerEntry(cmp, best);
            if (res != null) {
                best = res;
            }
        }
        cmp.add(3, best, false, false);
        for (int i = 0; i < groups.size(); i++) {
            rtrBgpGroup grp = groups.get(i);
            tabRoute<addrIP> wil = grp.getWilling(afi);
            tabRoute<addrIP> chg = grp.getChanged(afi);
            if ((wil == null) || (chg == null)) {
                needFull = true;
                continue;
            }
            tabRouteEntry<addrIP> ntry = best.copyBytes();
            boolean b = false;
            if (!wil.del(ntry)) {
                chg.add(3, ntry, false, false);
                b = true;
            }
            if ((afi == afiUni) || (afi == afiMlt)) {
                if (ntry.rouSrc == rtrBgpUtil.peerOriginate) {
                    grp.originatePrefix(ntry);
                } else if (grp.readvertPrefix(true, ntry)) {
                    continue;
                }
                if (tabRoute.addUpdatedEntry(3, wil, ntry, grp.roumapOut, grp.roupolOut, grp.prflstOut) < 1) {
                    continue;
                }
            } else {
                if (grp.readvertPrefix(false, ntry)) {
                    continue;
                }
                if (tabRoute.addUpdatedEntry(3, wil, ntry, grp.voumapOut, grp.voupolOut, null) < 1) {
                    continue;
                }
            }
            if (b) {
                continue;
            }
            chg.add(3, ntry, false, false);
        }
    }

    private void computeIncrUpdate(int afi, tabRoute<addrIP> chg, tabRoute<addrIP> cmp) {
        for (int i = chg.size() - 1; i >= 0; i--) {
            computeIncrEntry(afi, chg.del(i), cmp);
        }
    }

    private void computeIncrPurge(int ver, tabRoute<addrIP> chg) {
        for (int i = chg.size() - 1; i >= 0; i--) {
            tabRouteEntry<addrIP> ntry = chg.get(i);
            if (ntry.version >= ver) {
                continue;
            }
            chg.del(ntry);
        }
    }

    private boolean computeIncr() {
        long tim = bits.getTime();
        for (int i = 0; i < groups.size(); i++) {
            rtrBgpGroup grp = groups.get(i);
            if (grp.sendDefRou) {
                return true;
            }
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh nei = lstnNei.get(i);
            if (nei.softReconfig) {
                return true;
            }
            nei.setAccepted();
            if (nei.reachOld != nei.reachable) {
                return true;
            }
        }
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh nei = neighs.get(i);
            if (nei.softReconfig) {
                return true;
            }
            nei.setAccepted();
            if (nei.reachOld != nei.reachable) {
                return true;
            }
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " purge");
        }
        for (int i = 0; i < groups.size(); i++) {
            groups.get(i).minversion = compRound;
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            lstnNei.get(i).setGrpVer();
        }
        for (int i = 0; i < neighs.size(); i++) {
            neighs.get(i).setGrpVer();
        }
        for (int i = 0; i < groups.size(); i++) {
            rtrBgpGroup grp = groups.get(i);
            computeIncrPurge(grp.minversion, grp.chgUni);
            computeIncrPurge(grp.minversion, grp.chgMlt);
            computeIncrPurge(grp.minversion, grp.chgFlw);
            computeIncrPurge(grp.minversion, grp.chgVpnU);
            computeIncrPurge(grp.minversion, grp.chgVpnM);
            computeIncrPurge(grp.minversion, grp.chgVpnF);
            computeIncrPurge(grp.minversion, grp.chgVpls);
            computeIncrPurge(grp.minversion, grp.chgEvpn);
            computeIncrPurge(grp.minversion, grp.chgMdt);
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " changes");
        }
        computeIncrUpdate(afiUni, changedUni, routerComputed);
        computeIncrUpdate(afiMlt, changedMlt, computedMlt);
        computeIncrUpdate(afiFlw, changedFlw, computedFlw);
        computeIncrUpdate(afiVpnU, changedVpnU, computedVpnU);
        computeIncrUpdate(afiVpnM, changedVpnM, computedVpnM);
        computeIncrUpdate(afiVpnF, changedVpnF, computedVpnF);
        computeIncrUpdate(afiVpls, changedVpls, computedVpls);
        computeIncrUpdate(afiEvpn, changedEvpn, computedEvpn);
        computeIncrUpdate(afiMdt, changedMdt, computedMdt);
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " export");
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.doPeers(computedVpnU);
        }
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).doPeers(computedVpls);
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).doPeers(computedEvpn);
        }
        incrLast = bits.getTime();
        incrTime = (int) (incrLast - tim);
        incrCount++;
        return false;
    }

    private tabRouteEntry<addrIP> computeConquerEntry(tabRoute<addrIP> cmp, tabRouteEntry<addrIP> best) {
        if (best.nextHop == null) {
            return null;
        }
        tabRouteEntry<addrIP> old = cmp.find(best);
        if (old == null) {
            return null;
        }
        if (old.nextHop == null) {
            return null;
        }
        best = best.copyBytes();
        if (best.locPref < old.locPref) {
            best.locPref = old.locPref;
        }
        if (old.nextHop.compare(old.nextHop, best.nextHop) != 0) {
            best.locPref++;
        }
        return best;
    }

    private void computeConquerTable(tabRoute<addrIP> old, tabRoute<addrIP> cmp) {
        for (int i = 0; i < cmp.size(); i++) {
            tabRouteEntry<addrIP> ntry = cmp.get(i);
            ntry = computeConquerEntry(old, ntry);
            if (ntry == null) {
                continue;
            }
            cmp.add(3, ntry, false, false);
        }
    }

    public synchronized void routerCreateComputed() {
        if (debugger.rtrBgpEvnt) {
            logger.debug("create table");
        }
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " start");
        }
        int chg = changedUni.size() + changedMlt.size() + changedFlw.size() + changedVpnU.size() + changedVpnM.size() + changedVpnF.size() + changedVpls.size() + changedEvpn.size() + changedMdt.size();
        needFull |= chg > incrLimit;
        needFull |= oldAggr;
        if (routerAggregating.size() > 0) {
            needFull = true;
            oldAggr = true;
        } else {
            oldAggr = false;
        }
        if (needFull) {
            computeFull();
        } else if (computeIncr()) {
            computeFull();
        }
        changedCur = chg;
        changedTot += chg;
        if (debugger.rtrBgpComp) {
            logger.debug("round " + compRound + " done");
        }
        compRound++;
        fwdCore.routerChg(this);
    }

    public void routerCloseNow() {
        if (debugger.rtrBgpEvnt) {
            logger.debug("shutdown");
        }
        need2run = false;
        compute.wakeup();
        for (int i = 0; i < rpkis.size(); i++) {
            rtrBgpRpki ntry = rpkis.get(i);
            ntry.stopNow();
        }
        for (int i = 0; i < mons.size(); i++) {
            rtrBgpMon ntry = mons.get(i);
            ntry.stopNow();
        }
        for (int i = 0; i < dmps.size(); i++) {
            rtrBgpMrt ntry = dmps.get(i);
            ntry.stopNow();
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            ntry.stopNow();
            ntry.conn.closeNow();
        }
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            ntry.stopNow();
            ntry.conn.closeNow();
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.unregister2ip();
        }
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).doStop();
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).doStop();
        }
        tabLabel.release(evpnUni, 10);
        tabLabel.release(evpnMul, 10);
        fwdCore.routerDel(this);
    }

    public void routerGetHelp(userHelping l) {
        l.add("1 2   address-family              specify address families");
        rtrBgpParam.getAfiList(l, "2 2,.", "to use", true);
        l.add("1 2   local-as                    specify local as number");
        l.add("2 .     <num>                     autonomous system number");
        l.add("1 .   conquer                     conquer bestpath advertisements");
        l.add("1 2   incremental                 limit on incremental bestpath calculation");
        l.add("2 .     <num>                     maximum prefixes");
        l.add("1 2   router-id                   specify router id");
        l.add("2 .     <addr>                    router id");
        l.add("1 2   scantime                    scan time interval");
        l.add("2 .     <num>                     ms between scans");
        l.add("1 2   graceful-restart            graceful restart interval");
        l.add("2 .     <num>                     ms to recover");
        l.add("1 2   template                    specify template parameters");
        l.add("2 3     <name>                    name of template");
        rtrBgpParam.getParamHelp(l);
        l.add("1 2   nexthop                     specify next hop tracking parameter");
        l.add("2 3     route-map                 filter next hops");
        l.add("3 .       <name>                  name of route map");
        l.add("2 3     route-policy              filter next hops");
        l.add("3 .       <name>                  name of route policy");
        l.add("2 3     prefix-list               filter next hops");
        l.add("3 .       <name>                  name of prefix list");
        l.add("1 2   flowspec                    specify flowspec parameter");
        l.add("2 .     <name>                    name of policy map");
        l.add("1 2   neighbor                    specify neighbor parameters");
        l.add("2 3     <addr>                    address of peer");
        l.add("3 4       template                get configuration from template");
        l.add("4 .         <name>                name of source template");
        rtrBgpParam.getParamHelp(l);
        l.add("1 2   distance                    specify default distance");
        l.add("2 3     <num>                     external peer distance");
        l.add("3 4       <num>                   internal peer distance");
        l.add("4 .         <num>                 locally generated distance");
        l.add("1 2   listen                      passively listen for clients");
        l.add("2 3     <name>                    access list name");
        l.add("3 .       <name>                  template name");
        l.add("1 2   dump                        setup bgp dump file");
        l.add("2 3     <name>                    name of mrt");
        l.add("3 3,.     <file>                  name of file");
        l.add("1 2   monitor                     setup bgp monitor protocol server");
        l.add("2 3     <name>                    name of bmp");
        l.add("3 4       <name>                  proxy profile");
        l.add("4 5         <name>                hostname");
        l.add("5 .           <num>               port number");
        l.add("1 2   rpki                        setup resource public key infrastructure server");
        l.add("2 3     <name>                    name of bmp");
        l.add("3 4       <name>                  proxy profile");
        l.add("4 5         <name>                hostname");
        l.add("5 .           <num>               port number");
        l.add("1 2   afi-vrf                     select vrf to advertise");
        l.add("2 3     <vrf>                     name of routing table");
        l.add("3 .       enable                  enable processing");
        l.add("3 4       distance                set import distance");
        l.add("4 .         <num>                 distance");
        l.add("3 4       flowspec                specify flowspec parameter");
        l.add("4 .         <name>                name of policy map");
        cfgRtr.getRedistHelp(l, 2);
        l.add("1 2   afi-vpls                    select vpls to advertise");
        l.add("2 3     <id>                      vpls id in ASnum:IDnum format");
        l.add("3 4       bridge-group            enable processing");
        l.add("4 .         <name>                bridge group number");
        l.add("3 4       update-source           select source to advertise");
        l.add("4 .         <name>                name of interface");
        l.add("3 4       rd                      specify route distinguisher");
        l.add("4 .         <rd>                  rd in ASnum:IDnum format");
        l.add("3 4       rt-import               specify route target import");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       rt-export               specify route target export");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       rt-both                 specify route target");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       ve-id                   specify ve id");
        l.add("4 5         <num>                 ve id number");
        l.add("5 .           <num>               ve maximum number");
        l.add("1 2   afi-evpn                    select evpn to advertise");
        l.add("2 3     <id>                      evpn id");
        l.add("3 4       bridge-group            enable processing");
        l.add("4 .         <name>                bridge group number");
        l.add("3 4       bmac                    set backbone mac");
        l.add("4 .         <addr>                mac address");
        l.add("3 4       update-source           select source to advertise");
        l.add("4 .         <name>                name of interface");
        l.add("3 4       rd                      specify route distinguisher");
        l.add("4 .         <rd>                  rd in ASnum:IDnum format");
        l.add("3 4       rt-import               specify route target import");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       rt-export               specify route target export");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       rt-both                 specify route target");
        l.add("4 .         <rt>                  rt in ASnum:IDnum format");
        l.add("3 4       encapsulation           specify encapsulation to use");
        l.add("4 .         pbb                   pbb");
        l.add("4 .         vxlan                 vxlan");
        l.add("4 .         cmac                  cmac");
    }

    public void routerGetConfig(List<String> l, String beg, boolean filter) {
        l.add(beg + "local-as " + bits.num2str(localAs));
        l.add(beg + "router-id " + routerID);
        l.add(beg + "address-family" + rtrBgpParam.mask2string(addrFams));
        l.add(beg + "distance " + distantExt + " " + distantInt + " " + distantLoc);
        l.add(beg + "scantime " + scanTime);
        l.add(beg + "incremental " + incrLimit);
        l.add(beg + "graceful-restart " + restartTime);
        cmds.cfgLine(l, !conquer, beg, "conquer", "");
        cmds.cfgLine(l, nhtRoumap == null, beg, "nexthop route-map", "" + nhtRoumap);
        cmds.cfgLine(l, nhtRouplc == null, beg, "nexthop route-policy", "" + nhtRouplc);
        cmds.cfgLine(l, nhtPfxlst == null, beg, "nexthop prefix-list", "" + nhtPfxlst);
        cmds.cfgLine(l, flowSpec == null, beg, "flowspec", "" + flowSpec);
        for (int i = 0; i < mons.size(); i++) {
            mons.get(i).getConfig(l, beg);
        }
        for (int i = 0; i < dmps.size(); i++) {
            dmps.get(i).getConfig(l, beg);
        }
        for (int i = 0; i < rpkis.size(); i++) {
            rpkis.get(i).getConfig(l, beg);
        }
        for (int i = 0; i < temps.size(); i++) {
            temps.get(i).getConfig(l, beg, filter);
        }
        for (int i = 0; i < neighs.size(); i++) {
            neighs.get(i).getConfig(l, beg, filter);
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.getConfig(l, beg);
        }
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).getConfig(l, beg);
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).getConfig(l, beg);
        }
        if (lstnTmp == null) {
            l.add(beg + "no listen");
        } else {
            l.add(beg + "listen " + lstnAcl.listName + " " + lstnTmp.tempName);
        }
    }

    public boolean routerConfigure(cmds cmd) {
        String s = cmd.word();
        boolean negated = false;
        if (s.equals("no")) {
            s = cmd.word();
            negated = true;
        }
        if (s.equals("local-as")) {
            localAs = bits.str2num(cmd.word());
            return false;
        }
        if (s.equals("router-id")) {
            routerID.fromString(cmd.word());
            return false;
        }
        if (s.equals("address-family")) {
            addrFams = rtrBgpParam.string2mask(cmd);
            return false;
        }
        if (s.equals("distance")) {
            distantExt = bits.str2num(cmd.word());
            distantInt = bits.str2num(cmd.word());
            distantLoc = bits.str2num(cmd.word());
            return false;
        }
        if (s.equals("scantime")) {
            scanTime = bits.str2num(cmd.word());
            return false;
        }
        if (s.equals("incremental")) {
            incrLimit = bits.str2num(cmd.word());
            needFull = true;
            compute.wakeup();
            return false;
        }
        if (s.equals("conquer")) {
            conquer = !negated;
            needFull = true;
            compute.wakeup();
            return false;
        }
        if (s.equals("graceful-restart")) {
            restartTime = bits.str2num(cmd.word());
            return false;
        }
        if (s.equals("nexthop")) {
            s = cmd.word();
            if (s.equals("route-map")) {
                if (negated) {
                    nhtRoumap = null;
                    return false;
                }
                cfgRoump ntry = cfgAll.rtmpFind(cmd.word(), false);
                if (ntry == null) {
                    cmd.error("no such route map");
                    return false;
                }
                nhtRoumap = ntry.roumap;
                return false;
            }
            if (s.equals("route-policy")) {
                if (negated) {
                    nhtRouplc = null;
                    return false;
                }
                cfgRouplc ntry = cfgAll.rtplFind(cmd.word(), false);
                if (ntry == null) {
                    cmd.error("no such route map");
                    return false;
                }
                nhtRouplc = ntry.rouplc;
                return false;
            }
            if (s.equals("prefix-list")) {
                if (negated) {
                    nhtPfxlst = null;
                    return false;
                }
                cfgPrfxlst ntry = cfgAll.prfxFind(cmd.word(), false);
                if (ntry == null) {
                    cmd.error("no such prefix list");
                    return false;
                }
                nhtPfxlst = ntry.prflst;
                return false;
            }
            return true;
        }
        if (s.equals("flowspec")) {
            if (negated) {
                flowSpec = null;
                needFull = true;
                compute.wakeup();
                return false;
            }
            cfgPlymp ntry = cfgAll.plmpFind(cmd.word(), false);
            if (ntry == null) {
                cmd.error("no such policy map");
                return false;
            }
            flowSpec = ntry.plcmap;
            needFull = true;
            compute.wakeup();
            return false;
        }
        if (s.equals("afi-vrf")) {
            cfgVrf cfv = cfgAll.vrfFind(cmd.word(), false);
            if (cfv == null) {
                cmd.error("no such vrf");
                return false;
            }
            rtrBgpVrf cur = new rtrBgpVrf(this, cfv);
            s = cmd.word();
            if (s.equals("enable")) {
                rtrBgpVrf old = vrfs.find(cur);
                if (old != null) {
                    cur = old;
                    if (negated) {
                        old.doer.unregister2ip();
                        vrfs.del(old);
                        return false;
                    }
                    return false;
                }
                if (negated) {
                    return false;
                }
                cur.doer.register2ip();
                vrfs.put(cur);
                return false;
            }
            cur = vrfs.find(cur);
            if (cur == null) {
                cmd.error("vrf not enabled");
                return false;
            }
            if (s.equals("distance")) {
                cur.doer.distance = bits.str2num(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("flowspec")) {
                if (negated) {
                    cur.doer.flowSpec = null;
                    needFull = true;
                    compute.wakeup();
                    return false;
                }
                cfgPlymp ntry = cfgAll.plmpFind(cmd.word(), false);
                if (ntry == null) {
                    cmd.error("no such policy map");
                    return false;
                }
                cur.doer.flowSpec = ntry.plcmap;
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (cfgRtr.doCfgRedist(cur.doer, negated, s, cmd)) {
                cmd.badCmd();
            }
            needFull = true;
            compute.wakeup();
            return false;
        }
        if (s.equals("afi-vpls")) {
            rtrBgpVpls cur = new rtrBgpVpls(this);
            cur.id = tabRtrmapN.string2rd(cmd.word());
            s = cmd.word();
            if (s.equals("bridge-group")) {
                rtrBgpVpls old = vpls.del(cur);
                if (old != null) {
                    old.doStop();
                }
                if (negated) {
                    needFull = true;
                    compute.wakeup();
                    return false;
                }
                cur.bridge = cfgAll.brdgFind(cmd.word(), false);
                if (cur.bridge == null) {
                    cmd.error("no such bridge");
                    return false;
                }
                cur.rd = cur.id;
                cur.rtExp = cur.id;
                cur.rtImp = cur.id;
                vpls.add(cur);
                return false;
            }
            cur = vpls.find(cur);
            if (cur == null) {
                cmd.error("vpls not enabled");
                return false;
            }
            if (s.equals("rd")) {
                cur.rd = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-import")) {
                cur.rtImp = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-export")) {
                cur.rtExp = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-both")) {
                cur.rtExp = tabRtrmapN.string2rd(cmd.word());
                cur.rtImp = cur.rtExp;
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("ve-id")) {
                cur.veId = bits.str2num(cmd.word());
                cur.veMax = bits.str2num(cmd.word());
                if (negated) {
                    cur.veId = 0;
                    cur.veMax = 0;
                }
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("update-source")) {
                if (negated) {
                    cur.iface = null;
                } else {
                    cur.iface = cfgAll.ifcFind(cmd.word(), false);
                }
                needFull = true;
                compute.wakeup();
                return false;
            }
            return false;
        }
        if (s.equals("afi-evpn")) {
            rtrBgpEvpn cur = new rtrBgpEvpn(this);
            cur.id = bits.str2num(cmd.word());
            s = cmd.word();
            if (s.equals("bridge-group")) {
                rtrBgpEvpn old = evpn.del(cur);
                if (old != null) {
                    old.doStop();
                }
                if (negated) {
                    needFull = true;
                    compute.wakeup();
                    return false;
                }
                cur.bridge = cfgAll.brdgFind(cmd.word(), false);
                if (cur.bridge == null) {
                    cmd.error("no such bridge");
                    return false;
                }
                cur.bridge.bridgeHed.macRouter = cur;
                cur.rd = tabRtrmapN.string2rd(localAs + ":" + cur.id);
                cur.rtExp = cur.rd;
                cur.rtImp = cur.rd;
                cur.bbmac = addrMac.getRandom();
                cur.bcmac = ifcDot1ah.dstBmac4flood(cur.id);
                cur.encap = rtrBgpEvpn.encapType.pbb;
                evpn.add(cur);
                return false;
            }
            cur = evpn.find(cur);
            if (cur == null) {
                cmd.error("evpn not enabled");
                return false;
            }
            if (s.equals("bmac")) {
                cur.bbmac.fromString(cmd.word());
                return false;
            }
            if (s.equals("rd")) {
                cur.rd = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-import")) {
                cur.rtImp = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-export")) {
                cur.rtExp = tabRtrmapN.string2rd(cmd.word());
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("rt-both")) {
                cur.rtExp = tabRtrmapN.string2rd(cmd.word());
                cur.rtImp = cur.rtExp;
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("update-source")) {
                if (negated) {
                    cur.iface = null;
                } else {
                    cur.iface = cfgAll.ifcFind(cmd.word(), false);
                }
                needFull = true;
                compute.wakeup();
                return false;
            }
            if (s.equals("encapsulation")) {
                s = cmd.word();
                if (s.equals("pbb")) {
                    cur.encap = rtrBgpEvpn.encapType.pbb;
                }
                if (s.equals("vxlan")) {
                    cur.encap = rtrBgpEvpn.encapType.vxlan;
                }
                if (s.equals("cmac")) {
                    cur.encap = rtrBgpEvpn.encapType.cmac;
                }
                needFull = true;
                compute.wakeup();
                return false;
            }
            return false;
        }
        if (s.equals("dump")) {
            rtrBgpMrt dmp = new rtrBgpMrt();
            dmp.dumpName = cmd.word();
            if (negated) {
                dmp = dmps.del(dmp);
                if (dmp == null) {
                    return false;
                }
                dmp.stopNow();
                return false;
            }
            dmp.fileName = cmd.getRemaining();
            dmp.startNow();
            dmps.add(dmp);
            return false;
        }
        if (s.equals("monitor")) {
            rtrBgpMon mon = new rtrBgpMon();
            mon.monName = cmd.word();
            if (negated) {
                mon = mons.del(mon);
                if (mon == null) {
                    return false;
                }
                mon.stopNow();
                return false;
            }
            cfgProxy prx = cfgAll.proxyFind(cmd.word(), false);
            if (prx == null) {
                cmd.error("no such proxy");
                return false;
            }
            mon.proxy = prx.proxy;
            mon.server = cmd.word();
            mon.port = bits.str2num(cmd.word());
            mon.startNow();
            mons.add(mon);
            return false;
        }
        if (s.equals("rpki")) {
            rtrBgpRpki rpki = new rtrBgpRpki(this);
            rpki.rpkiName = cmd.word();
            if (negated) {
                rpki = rpkis.del(rpki);
                if (rpki == null) {
                    return false;
                }
                rpki.stopNow();
                return false;
            }
            cfgProxy prx = cfgAll.proxyFind(cmd.word(), false);
            if (prx == null) {
                cmd.error("no such proxy");
                return false;
            }
            rpki.proxy = prx.proxy;
            rpki.server = cmd.word();
            rpki.port = bits.str2num(cmd.word());
            rpki.startNow();
            rpkis.add(rpki);
            return false;
        }
        if (s.equals("listen")) {
            if (negated) {
                lstnAcl = null;
                lstnTmp = null;
                tcpCore.listenStop(null, port, null, 0, 0);
                return false;
            }
            cfgAceslst acl = cfgAll.aclsFind(cmd.word(), false);
            if (acl == null) {
                cmd.error("no such acl");
                return true;
            }
            lstnTmp = findTemp(cmd.word());
            if (lstnTmp == null) {
                cmd.error("no such template");
                return true;
            }
            lstnAcl = acl.aceslst;
            tcpCore.streamListen(this, new pipeLine(lstnTmp.bufferSize, false), null, port, null, 0, 0, "bgp", lstnTmp.passwd, lstnTmp.ttlSecurity);
            return false;
        }
        if (s.equals("template")) {
            rtrBgpTemp ntry = new rtrBgpTemp(this);
            ntry.tempName = cmd.word();
            rtrBgpTemp old = temps.add(ntry);
            if (old != null) {
                ntry = old;
            }
            negated = ntry.setParamCfg(cmd, negated);
            if (ntry.remoteAs != 0) {
                return negated;
            }
            temps.del(ntry);
            return negated;
        }
        if (!s.equals("neighbor")) {
            return true;
        }
        rtrBgpNeigh ntry = new rtrBgpNeigh(this);
        if (ntry.peerAddr.fromString(cmd.word())) {
            cmd.error("bad address");
            return false;
        }
        rtrBgpNeigh old = neighs.add(ntry);
        if (old == null) {
            ntry.startNow();
        } else {
            ntry = old;
        }
        negated = ntry.setParamCfg(cmd, negated);
        ntry.updatePeer();
        needFull = true;
        compute.wakeup();
        if (ntry.remoteAs != 0) {
            return negated;
        }
        ntry.stopNow();
        neighs.del(ntry);
        return negated;
    }

    /**
     * template configuration
     *
     * @param temp template interface
     * @param cmd command to parse
     * @param negated negated
     */
    public void templateConfig(rtrBgpTemp temp, String cmd, boolean negated) {
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            if (ntry.template == null) {
                continue;
            }
            if (!temp.tempName.equals(ntry.template.tempName)) {
                continue;
            }
            ntry.setParamCfg(new cmds("template", cmd), negated);
        }
    }

    /**
     * list neighbors
     *
     * @return list of neighbors
     */
    public userFormat showRpkiNei() {
        userFormat l = new userFormat("|", "learn|neighbor");
        for (int i = 0; i < rpkis.size(); i++) {
            rtrBgpRpki ntry = rpkis.get(i);
            l.add(ntry.table.size() + "|" + ntry.server);
        }
        return l;
    }

    /**
     * list neighbor summary
     *
     * @param safi safi to query
     * @return list of neighbors
     */
    public userFormat showNeighs(int safi) {
        userFormat l = new userFormat("|", "as|learn|accept|will|done|neighbor|uptime");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNeighs(safi));
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNeighs(safi));
        }
        return l;
    }

    /**
     * list neighbor summary
     *
     * @return list of neighbors
     */
    public userFormat showNeighs() {
        userFormat l = new userFormat("|", "as|afis|neighbor|uptime");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNeighs());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNeighs());
        }
        return l;
    }

    /**
     * list group summary
     *
     * @return list of groups
     */
    public userFormat showGroups() {
        userFormat l = new userFormat("|", "as|group|neighbor|uptime");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showGroups());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showGroups());
        }
        return l;
    }

    /**
     * list nexthop summary
     *
     * @return list of nexthops
     */
    public userFormat showNxtHps() {
        userFormat l = new userFormat("|", "as|reach|chg|num|neighbor|uptime");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNxtHps());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showNxtHps());
        }
        return l;
    }

    /**
     * list graceful restart
     *
     * @return list of nexthops
     */
    public userFormat showGrcRst() {
        userFormat l = new userFormat("|", "as|rx|tx|neighbor");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showGrcRst());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showGrcRst());
        }
        return l;
    }

    /**
     * list additional path
     *
     * @return list of nexthops
     */
    public userFormat showAddPth() {
        userFormat l = new userFormat("|", "as|rx|tx|neighbor");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showAddPth());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showAddPth());
        }
        return l;
    }

    /**
     * list router id
     *
     * @return list of nexthops
     */
    public userFormat showRtrId() {
        userFormat l = new userFormat("|", "as|router|wideas|refresh|neighbor");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showRtrId());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showRtrId());
        }
        return l;
    }

    /**
     * list buffers
     *
     * @return list of nexthops
     */
    public userFormat showBuffer() {
        userFormat l = new userFormat("|", "as|buf|conn|neighbor");
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh ntry = neighs.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showBuffer());
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh ntry = lstnNei.get(i);
            if (ntry == null) {
                continue;
            }
            l.add(ntry.showBuffer());
        }
        return l;
    }

    /**
     * find peer
     *
     * @param adr address to find
     * @return neighbor, null if not found
     */
    public rtrBgpNeigh findPeer(addrIP adr) {
        rtrBgpNeigh ntry = new rtrBgpNeigh(this);
        ntry.peerAddr = adr.copyBytes();
        rtrBgpNeigh res = neighs.find(ntry);
        if (res != null) {
            return res;
        }
        return lstnNei.find(ntry);
    }

    /**
     * find group
     *
     * @param num number of group
     * @return group, null if not found
     */
    public rtrBgpGroup findGroup(int num) {
        if (num < 0) {
            return null;
        }
        if (num >= groups.size()) {
            return null;
        }
        return groups.get(num);
    }

    /**
     * find template
     *
     * @param nam name to find
     * @return template, null if not found
     */
    public rtrBgpTemp findTemp(String nam) {
        rtrBgpTemp ntry = new rtrBgpTemp(this);
        ntry.tempName = "" + nam;
        return temps.find(ntry);
    }

    /**
     * get database
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getDatabase(int safi) {
        if (safi == afiUni) {
            return routerComputed;
        }
        if (safi == afiLab) {
            return routerComputed;
        }
        if (safi == afiMlt) {
            return computedMlt;
        }
        if (safi == afiFlw) {
            return computedFlw;
        }
        if (safi == afiVpnU) {
            return computedVpnU;
        }
        if (safi == afiVpnM) {
            return computedVpnM;
        }
        if (safi == afiVpnF) {
            return computedVpnF;
        }
        if (safi == afiVpls) {
            return computedVpls;
        }
        if (safi == afiEvpn) {
            return computedEvpn;
        }
        if (safi == afiMdt) {
            return computedMdt;
        }
        return null;
    }

    /**
     * get changed
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getChanged(int safi) {
        if (safi == afiUni) {
            return changedUni;
        }
        if (safi == afiLab) {
            return changedUni;
        }
        if (safi == afiMlt) {
            return changedMlt;
        }
        if (safi == afiFlw) {
            return changedFlw;
        }
        if (safi == afiVpnU) {
            return changedVpnU;
        }
        if (safi == afiVpnM) {
            return changedVpnM;
        }
        if (safi == afiVpnF) {
            return changedVpnF;
        }
        if (safi == afiVpls) {
            return changedVpls;
        }
        if (safi == afiEvpn) {
            return changedEvpn;
        }
        if (safi == afiMdt) {
            return changedMdt;
        }
        return null;
    }

    /**
     * set validity
     *
     * @param ntry entry to update
     */
    protected void setValidity(tabRouteEntry<addrIP> ntry) {
        if (rpkis.size() < 1) {
            return;
        }
        tabRouteEntry<addrIP> r = computedRpki.route(ntry.prefix.broadcast);
        if (r == null) {
            ntry.validity = 2;
            return;
        }
        if (ntry.pathSeq == null) {
            ntry.validity = 3;
            return;
        }
        if (ntry.pathSeq.size() < 1) {
            ntry.validity = 3;
            return;
        }
        if (ntry.pathSeq.get(ntry.pathSeq.size() - 1) != r.rouSrc) {
            ntry.validity = 3;
            return;
        }
        ntry.validity = 1;
    }

    public int routerNeighCount() {
        return neighs.size() + lstnNei.size();
    }

    public void routerNeighList(tabRoute<addrIP> tab) {
        for (int i = 0; i < neighs.size(); i++) {
            rtrBgpNeigh nei = neighs.get(i);
            if (nei == null) {
                continue;
            }
            tabRouteEntry<addrIP> ntry = new tabRouteEntry<addrIP>();
            ntry.prefix = new addrPrefix<addrIP>(nei.peerAddr, addrIP.size * 8);
            tabRoute.addUpdatedEntry(2, tab, ntry, null, null, routerAutoMesh);
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            rtrBgpNeigh nei = lstnNei.get(i);
            if (nei == null) {
                continue;
            }
            tabRouteEntry<addrIP> ntry = new tabRouteEntry<addrIP>();
            ntry.prefix = new addrPrefix<addrIP>(nei.peerAddr, addrIP.size * 8);
            tabRoute.addUpdatedEntry(2, tab, ntry, null, null, routerAutoMesh);
        }
        for (int i = 0; i < vrfs.size(); i++) {
            vrfs.get(i).doer.getPeerList(tab);
        }
        for (int i = 0; i < vpls.size(); i++) {
            vpls.get(i).getPeerList(tab);
        }
        for (int i = 0; i < evpn.size(); i++) {
            evpn.get(i).getPeerList(tab);
        }
    }

    public int routerIfaceCount() {
        return 0;
    }

    /**
     * get all routes
     *
     * @param safi safi to query
     * @param prf prefix to find
     * @return list of routes
     */
    public List<String> getAllRoutes(int safi, tabRouteEntry<addrIP> prf) {
        List<String> lst = new ArrayList<String>();
        for (int i = 0; i < neighs.size(); i++) {
            getAllRoutes(lst, neighs.get(i), safi, prf);
        }
        for (int i = 0; i < lstnNei.size(); i++) {
            getAllRoutes(lst, lstnNei.get(i), safi, prf);
        }
        return lst;
    }

    private void getAllRoutes(List<String> lst, rtrBgpNeigh nei, int safi, tabRouteEntry<addrIP> prf) {
        if (nei == null) {
            return;
        }
        tabRoute<addrIP> tab = nei.conn.getLearned(safi);
        if (tab == null) {
            return;
        }
        tabRouteEntry<addrIP> res = tab.find(prf);
        if (res == null) {
            return;
        }
        lst.add("*************************** " + nei.peerAddr);
        lst.addAll(res.fullDump(fwdCore));
    }

    /**
     * get all routes
     *
     * @return list of statistics
     */
    public userFormat getBestpath() {
        userFormat l = new userFormat("|", "category|value");
        l.add("version|" + compRound);
        l.add("full run|" + fullCount + " times");
        l.add("full last|" + bits.time2str(cfgAll.timeZoneName, fullLast + cfgAll.timeServerOffset, 3) + " (" + bits.timePast(fullLast) + " ago)");
        l.add("full time|" + fullTime + " ms");
        l.add("incr run|" + incrCount + " times");
        l.add("incr last|" + bits.time2str(cfgAll.timeZoneName, incrLast + cfgAll.timeServerOffset, 3) + " (" + bits.timePast(incrLast) + " ago)");
        l.add("incr time|" + incrTime + " ms");
        l.add("changes|" + changedCur + ", total=" + changedTot);
        l.add("rpki table|" + computedRpki.size());
        l.add("unicast table|" + routerComputed.size() + ", list=" + changedUni.size());
        l.add("multicast table|" + computedMlt.size() + ", list=" + changedMlt.size());
        l.add("flowspec table|" + computedFlw.size() + ", list=" + changedFlw.size());
        l.add("vpnuni table|" + computedVpnU.size() + ", list=" + changedVpnU.size());
        l.add("vpnmlt table|" + computedVpnM.size() + ", list=" + changedVpnM.size());
        l.add("vpnflw table|" + computedVpnF.size() + ", list=" + changedVpnF.size());
        l.add("vpls table|" + computedVpls.size() + ", list=" + changedVpls.size());
        l.add("evpn table|" + computedEvpn.size() + ", list=" + changedEvpn.size());
        l.add("mdt table|" + computedMdt.size() + ", list=" + changedMdt.size());
        return l;
    }

}
